;;; emit.scm  --  Jens Axel Sgaard

(module emit mzscheme
  (provide emit)
  
  (require (lib "trace.ss"    "mzlib")
           (lib "pretty.ss"   "mzlib")
           (lib "moddep.ss"   "syntax")
           (lib "kerncase.ss" "syntax")
           "module-dependencies.scm")
  
  
  ;top-level-expr is one of
  ;  general-top-level-expr
  ;  (module identifier name (#%plain-module-begin module-level-expr ...))
  ;  (begin top-level-expr ...)
  
  ;module-level-expr is one of
  ;  general-top-level-expr
  ;  (provide provide-spec ...)
  ;  (begin module-level-expr ...)
  
  ;general-top-level-expr is one of
  ;  expr
  ;  (define-values (variable ...) expr)
  ;  (define-syntaxes (variable ...) expr)
  ;  (require require-spec ...)
  ;  (require-for-syntax require-spec ...)
  
  (define (emit stx)
    (emit-top-level-expr 0 stx))
  
  (define (emit-top-level-expr rp stx)
    (syntax-case* stx (module begin #%plain-module-begin) module-identifier=?
      [(module identifier name (#%plain-module-begin module-level-expr ...))
       #`(begin 
           '(module identifier name)
           #,@(emit-module-level-exprs rp #'(module-level-expr ...)))]
      [(begin top-level-expr ...)
       #`(begin #,@(emit-top-level-exprs rp #'(top-level-expr ...)))]
      [_
       (emit-general-top-level-expr rp stx)]))
  
  (define (emit-top-level-exprs rp stx)
    #`(#,@(map (lambda (e) emit-top-level-expr rp e)
               (syntax->list stx))))
  
  (define (emit-module-level-expr rp stx)
    (syntax-case stx (provide begin)
      [(provide provide-spec ...)
       #''(provide provide-spec ...)]
      [(begin module-level-expr ...)
       #`(begin #,(emit-module-level-exprs rp #'(module-level-expr ...)))]
      [_
       (emit-general-top-level-expr rp stx)]))
  
  (define (emit-module-level-exprs rp stx)
    #`(#,@(map (lambda (e) (emit-module-level-expr rp e))
               (syntax->list stx))))
  

;  (define (emit-general-top-level-expr rp stx)
;    (syntax-case stx (define-values define-syntaxes require require-for-syntax)    
;      [(define-values (variable ...) expr)
;       #`(define-values (variable ...) #,(emit-expr rp #'expr))]
;      [(define-syntaxes (variable ...) expr)
;       #`(define-syntaxes (variable ...) #,(emit-expr (+ rp 1) #'expr))]
;      [(require require-spec ...)
;       #`'(require require-spec ...)]
;      [(require-for-syntax require-spec ...)
;       #`'(require-for-syntax require-spec ...)]
;      [_
;       (emit-expr rp stx)]))

  (define (emit-general-top-level-expr rp stx)
    (syntax-case stx (define-values define-syntaxes require require-for-syntax)    
      [(require require-spec ...)
       #`'(require require-spec ...)]
      [(require-for-syntax require-spec ...)
       #`'(require-for-syntax require-spec ...)]
      [_
       stx]))


  
;  (define (emit-expr rp stx)
;    (syntax-case* stx (lambda case-lambda if begin begin0 
;                        let-values letrec-values set! quote quote-syntax
;                        with-continuation-mark #%app #%datum #%top)  (if (= rp 0)
;                                                                         module-identifier=?
;                                                                         module-transformer-identifier=?)
;      [(lambda formals expr0 expr ...)
;       #`(lambda formals #,@(emit-exprs rp #'(expr0 expr ...)))]
;      [(case-lambda (formals expr0 expr ...) ...)
;       #`(case-lambda #,@(emit-case-lambda-clauses rp #'((formals expr0 expr ...) ...)))]
;      [(if expr0 expr1)
;       #`(if #,(emit-expr rp #'expr0) #,(emit-expr rp #'expr1))]
;      [(if expr0 expr1 expr2)
;       #`(if #,(emit-expr rp #'expr0) #,(emit-expr rp #'expr1) #,(emit-expr rp #'expr2))]
;      [(begin expr0 expr ...)
;       #`(begin #,@(emit-exprs rp #'(expr0 expr ...)))]
;      [(begin0 expr0 expr ...)
;       #`(begin0 #,@(emit-exprs rp #'(expr0 expr ...)))]
;      [(let-values (((variable ...) expr) ...) expr0 expr1 ...)
;       #`(let-values #,(emit-let-values-clauses rp #'(((variable ...) expr) ...)) 
;           #,@(emit-exprs rp #'(expr0 expr1 ...)))]
;      [(letrec-values (((variable ...) expr) ...) expr0 expr1 ...)
;       #`(letrec-values  #,(emit-letrec-values-clauses rp #'(((variable ...) expr) ...)) 
;           #,@(emit-exprs rp #'(expr0 expr1 ...)))]
;      [(set! variable expr)
;       #`(set! variable #,(emit-expr rp #'expr))]
;      [(quote datum)
;       #'(quote datum)]
;      ; quote-syntax is the same as syntax without pattern variables
;      [(quote-syntax datum)
;       #`(quote-syntax #,(emit-quote-syntax rp #'datum))]
;      [(with-continuation-mark expr0 expr1 expr2)
;       #`(with-continuation-mark 
;             #,(emit-expr rp #'expr0) #,(emit-expr rp #'expr1) #,(emit-expr rp #'expr))]
;      [(#%app expr0 expr1 ...)
;       #`(#,@(emit-exprs rp #'(expr0 expr1 ...)))]
;      [(#%datum . datum)
;       #'datum]
;      [(#%top . variable)
;       #'variable]
;      [variable
;       (identifier? #'variable)
;       #'variable]
;      [failed
;       (begin 
;         (display "Hprm!") (display #'failed) (newline)
;         (error "huh?, got: " #'foo))]))
;  
;  (define (emit-let-values-clauses rp clauses-stx)
;    (define (emit-clause clause)
;      (syntax-case clause ()
;        [((variable ...) expr)
;         #`((variable ...) #,(emit-expr rp #'expr))]))
;    
;    (datum->syntax-object clauses-stx (map emit-clause (syntax->list clauses-stx))))
;  
;  (define emit-letrec-values-clauses emit-let-values-clauses)
;  
;  (define (emit-case-lambda-clauses rp clauses-stx)
;    (define (emit-clause clause)
;      (syntax-case clause ()
;        [((variable ...)       expr0 expr ...)   #`((variable ...)         #,@(emit-exprs rp #'(expr0 expr ...)))]
;        [((variable ... . var) expr0 expr ...)   #`((variable ... . var)   #,@(emit-exprs rp #'(expr0 expr ...)))]
;        [(variable             expr0 expr ...)   #`(variable               #,@(emit-exprs rp #'(expr0 expr ...)))]))
;    
;    (datum->syntax-object clauses-stx (map emit-clause (syntax->list clauses-stx))))
;  
;  
;  
;  (define (emit-quote-syntax rp datum)
;    ; TODO: think
;    datum)
;  
;  
;  (define (emit-exprs rp stx)
;    (datum->syntax-object stx (map (lambda (s) (emit-expr rp s))
;                                   (syntax->list stx)))))
;

  )