; module-dependencies.scm  --  Jens Axel Søgaard

(module module-dependencies mzscheme
  (provide process-dependencies
           expanded-module->direct-requires
           expanded-module->direct-for-syntax-requires
           expanded-module->module-name
           module-path-index->list)
  
  (require (lib "list.ss"   "mzlib")
           (lib "moddep.ss" "syntax"))
  
  (define *VERBOSE* #f)

  
  (define (expanded-module->direct-requires stx)
    (syntax-property stx 'module-direct-requires))
  
  (define (expanded-module->direct-for-syntax-requires stx)
    (syntax-property stx 'module-direct-for-syntax-requires))
  
  (define (complete-file-path->directory+filename path)
    (let-values ([(base name must-be-dir?) (split-path path)])
      (values base name)))
  
  (define (symbol-append . syms)
    (string->symbol (apply string-append (map symbol->string syms))))
  
;  (define (dependencies phase module-path)
;    (define (remove-kernel-modules l) (filter (lambda (i) (not (eqv? i 'mzscheme))) l))
;    
;    ; reparameterize current-load-relative-directory, if module-path is a complete path
;    (let-values ([(rel-dir module-path) 
;                  (cond
;                    [(complete-path? module-path)  (complete-file-path->directory+filename module-path)]
;                    [else                          (values (current-load-relative-directory) module-path)])])
;      (parameterize ([current-load-relative-directory rel-dir])
;        (let loop ([phase phase]
;                   [path  (resolve-module-path module-path #f)])
;          (cond
;            [(not path) '()]
;            [else
;             ; read the module
;             (let-values (((base name must-be-dir?) (split-path path)))
;               (let ([module-source (read-syntax name (open-input-file path))])
;                 ; expand it 
;                 (parameterize ([current-load-relative-directory base])
;                   (let* ([expanded-module (expand module-source)]
;                          [module-name     (expanded-module->module-name expanded-module)])
;                     ; find dependencies
;                     (let ([direct-for-syntax-requires (expanded-module->direct-for-syntax-requires expanded-module)]
;                           [direct-requires            (expanded-module->direct-requires expanded-module)])
;                       (cond
;                         ; the builtin module 'mzsheme has no dependencies
;                         [(eqv? module-name 'mzscheme)  (begin
;                                                          (display 'mzscheme) (newline)
;                                                          '())]
;                         [else 
;                          ; format the information
;                          (begin0 
;                            (cons
;                             `(,(symbol-append module-name ': (string->symbol (number->string phase))) 
;                                -->
;                                (requires: ,@(apply append (map (lambda (index) (module-path-index->list index path))
;                                                                direct-requires)))
;                                (for-syntax-requires: ,@(apply append (map (lambda (index) 
;                                                                             (module-path-index->list index path))
;                                                                           direct-for-syntax-requires))))
;                             (append (apply append (map (lambda (index)
;                                                          (loop (+ phase 1) (resolve-module-path-index index path)))
;                                                        (remove-kernel-modules direct-for-syntax-requires)))
;                                     (apply append (map (lambda (index)
;                                                          (loop phase (resolve-module-path-index index path)))
;                                                        (remove-kernel-modules direct-requires)))))
;                            (display (symbol-append module-name ': (string->symbol (number->string phase))))
;                            (newline))
;                          ]))))))])))))
  
  
  ; process-dependencies : phase module-path (phase module-name expanded-module -> ?) -> (void)
  ;   Call process on all (recursively) required or required-for-syntax
  ;   modules. 
  ;   E.g.  (define (process phase module-name expanded-module)  
  ;           (display (symbol-append module-name ': (string->symbol (number->string phase))))
  ;           (newline))
  (define (process-dependencies phase module-path process)
    (define processed-modules-ht  (make-hash-table 'equal))
    (define (process-kernel-module phase module-name expanded-module)
      (cond
        [(hash-table-get processed-modules-ht (list phase module-name ) (lambda () #f))
         (when *VERBOSE*
           (display "skipping: ") (display (list phase module-name)) (newline))]
        [else
         (hash-table-put! processed-modules-ht (list phase module-name) 'done)
         (process phase module-name expanded-module)])) 
    ; TODO: Memoize already seen modules, so they are only expanded once
    ; reparameterize current-load-relative-directory, if module-path is a complete path
    (let-values ([(rel-dir module-path) 
                  (cond
                    [(complete-path? module-path)  (complete-file-path->directory+filename module-path)]
                    [else                          (values (current-load-relative-directory) module-path)])])
      (parameterize ([current-load-relative-directory rel-dir])
        (let loop ([phase phase]
                   [path  (resolve-module-path module-path #f)])   ; Note: path is a complete path
          ; check whether we have seen the module before
          (cond
            [(hash-table-get processed-modules-ht (list path phase) (lambda () #f))
             (when *VERBOSE*
               (display "skipping: ") (display (list phase path)) (newline))]
            [else
             (hash-table-put! processed-modules-ht (list path phase) 'done)
             ; read the module
             (let-values (((base name must-be-dir?) (split-path path)))
               (call-with-input-file path 
                 (lambda (input-port)
                   (let ([module-source (read-syntax name input-port)])
                     ; expand it 
                     (parameterize ([current-load-relative-directory base])
                       (let* ([expanded-module (expand module-source)]   ; <===== MEMOIZE THIS
                              [module-name     (expanded-module->module-name expanded-module)])
                         ; find dependencies
                         (let ([direct-for-syntax-requires (expanded-module->direct-for-syntax-requires expanded-module)]
                               [direct-requires            (expanded-module->direct-requires expanded-module)])
                           (for-each (lambda (index)
                                       (if (symbol? index)  
                                           (process-kernel-module (+ phase 1) index expanded-module)
                                           (loop (+ phase 1) (resolve-module-path-index index path))))
                                     direct-for-syntax-requires)
                           (for-each (lambda (index)
                                       (if (symbol? index)
                                           (process-kernel-module phase index expanded-module)
                                           (loop phase (resolve-module-path-index index path))))
                                     direct-requires)
                           (process phase module-name expanded-module))))))))])))))
  
  
  (define (expanded-module->module-name stx)
    (syntax-case stx (module #%plain-module-begin)
      [(module identifier name (#%plain-module-begin module-level-expr ...))
       (syntax-e #'identifier)]
      [_
       (error #f "module expected, got: " stx)]))
  
  
  (define (module-path-index->list index self)
    (cond
      [(symbol? index) (list index)]
      [else            (let-values ([(path relative-to) (module-path-index-split index)])
                         (cond
                           [(eqv? path #f)  
                            ; relative to self (implies relative-to is #f too)
                            '() ;(list self)
                            ]
                           [else
                            (cons path (cond
                                         [(module-path-index? relative-to)
                                          (module-path-index->list relative-to self)]
                                         [(symbol? relative-to)
                                          (list relative-to)]
                                         [(eqv? relative-to #f)
                                          '(*top-level*)]
                                         [else
                                          (error "huh?")]))]))]))
  )
 

;;;
;;;  TEST
;;;


;(require module-dependencies)
;
;(define a-path "c:/Documents and Settings/js/Skrivebord/DjMeta/a.scm")
;(define b-path "c:/Documents and Settings/js/Skrivebord/DjMeta/b.scm")
;(define c-path "c:/Documents and Settings/js/Skrivebord/DjMeta/c.scm")
;
;  
;;'a
;;(pretty-print (dependencies 0 "a.scm"))
;;'b
;;(pretty-print (dependencies 0 b-path))
;;'c
;;(pretty-print (dependencies 0 c-path))
;
;(define (process phase module-name expanded-module) 
;  (display (symbol-append module-name ': (string->symbol (number->string phase))))
;  (newline))
;
;;'a
;;(process-dependencies 0 "a.scm" process)
;;'b
;;(process-dependencies 0 b-path process)
;'c
;(process-dependencies 0 c-path process)
