(require (lib "trace.ss"    "mzlib")
         (lib "pretty.ss"   "mzlib")
         (lib "moddep.ss"   "syntax")
         (lib "kerncase.ss" "syntax")
         "module-dependencies.scm"
         "emit.scm")

(define *MZMODE* #f)

; filename-without-extension : path-or-string -> string
(define (filename-without-extension filename)
  (let ((filename (if (path? filename)
                      (path->string filename)
                      filename)))
    (let*([extension (filename-extension filename)]
          [extension (cond
                       [(path? extension)  (path->string extension)]
                       [(bytes? extension) (bytes->string/utf-8 extension)]
                       [else               extension])])
      (if extension
          (substring filename 0 (- (string-length filename) 
                                   1
                                   (string-length extension)))
          filename))))

(define (module-name-resolve module-path-index n)
  (if (module-path-index? module-path-index)
      (filename-without-extension
          (file-name-from-path 
           (resolve-module-path-index module-path-index (symbol->string n))))
      module-path-index))

;(define (expanded-module->module-name stx)
;  (syntax-case stx (module #%plain-module-begin)
;    [(module identifier name (#%plain-module-begin module-level-expr ...))
;     (syntax-e #'identifier)]
;    [_
;     (error #f "module expected, got: " stx)]))

;(define (module-path-index->symbol module-path-index)
;  (map ->symbol 
;       (intersperse '$ (module-path-index->list module-path-index #f))))


; symbol-append : (list symbol) -> symbol
(define (symbol-append . syms)
  (string->symbol (apply string-append (map symbol->string syms))))

; ->symbol : object -> symbol
(define (->symbol o)
  (cond
    [(symbol? o) o]
    [(string? o) (string->symbol o)]
    [(number? o) (->symbol (number->string o))]
    [(path? o)   (string->symbol (path->string o))]
    [else        (error o)]))

; generate-identifier-name : phase relative-phase name id-stx -> id-stx
;   generate a canonical name for id-stx consisting of the
;   modulename name, (+ phase relative-phase) and the name of id-stx.
(define (generate-identifier-name p rp n id-stx)
  ;  The phase p is the phase the current module is expanded in. During this 
  ;  expansion, some expressions are expanded in phase (+ p rp) [rp = relative phase]
  ; (display "generate-identifier-name> ") (display (list p rp n id-stx)) (newline)
  (let-values ([(rp result)
                (let ((id-binding             (identifier-binding id-stx))
                      (id-transformer-binding (identifier-transformer-binding id-stx)))
                  (cond
                    [(and (= rp 0) id-binding)             (values 0 id-binding)]
                    [(and (= rp 1) id-transformer-binding) (values 1 id-transformer-binding)]
                    [(and (= rp 1) id-binding)             (values 0 id-binding)]
                    [(> rp 1)                              (error "Now this is odd!")]
                    [else                                  (values rp #f)]))])
    (cond
      ; bound to something else than a top-level or module variable
      [(eqv? result 'lexical) id-stx]
      ; module-defined variable
      [(list? result)         (let ((source-mod         (list-ref result 0))
                                    (source-id          (list-ref result 1))
                                    (nominal-source-mod (list-ref result 2))
                                    (nominal-source-id  (list-ref result 3)))
                                (cond
                                  [(eqv? source-mod '#%kernel)
                                   (datum->syntax-object id-stx source-id)]
                                  [else
                                   ; (for-each display (list "gen> " (module-path-index->list source-mod #f) "\n"))
                                   (datum->syntax-object id-stx
                                                         (symbol-append
                                                          (->symbol (module-name-resolve source-mod n)) ':
                                                          (->symbol (number->string (+ rp p))) ':
                                                          source-id))]))]
      ; not bound, or top-level
      [else                     (begin
                                  ; (for-each display (list "gen> " id-stx "\n"))
                                  id-stx)])))

;(define (generate-identifier-name p id-stx)
;  (display id-stx) (newline)
;  (let-values ([(p result)
;                (let ((id-binding             (identifier-binding id-stx))
;                      (id-transformer-binding (identifier-transformer-binding id-stx)))
;                  (cond
;                    [(and (= p 0) id-binding)             (values 0 id-binding)]
;                    [(and (= p 1) id-transformer-binding) (values 1 id-transformer-binding)]
;                    [(and (= p 1) id-binding)             (values 0 id-binding)]
;                    [(> p 1)                              (error "Now this is odd!")]
;                    [else                                 (values p #f)]))])
;    (cond
;      ; bound to something else than a top-level or module variable
;      [(eqv? result 'lexical) id-stx]
;      ; module-defined variable
;      [(list? result)         (let ((source-mod         (list-ref result 0))
;                                    (source-id          (list-ref result 1))
;                                    (nominal-source-mod (list-ref result 2))
;                                    (nominal-source-id  (list-ref result 3)))
;                                (cond
;                                  [(eqv? source-mod '#%kernel)
;                                   (datum->syntax-object id-stx source-id)]
;                                  [else
;                                   (for-each display (list "gen> " (module-path-index->list source-mod #f) "\n"))
;                                   (datum->syntax-object id-stx
;                                                         (symbol-append
;                                                          ; (->symbol (current-module-to-be-reexpanded)) ':
;                                                          (->symbol (module-name-resolve source-mod)) ':
;                                                          (->symbol (number->string p)) ':
;                                                          source-id))]))]
;      ; not bound, or top-level
;      [else                     (begin
;                                  (for-each display (list "gen> " id-stx "\n"))
;                                  id-stx)])))


; top-level-expr is one of
;   general-top-level-expr
;   (module identifier name (#%plain-module-begin module-level-expr ...))
;   (begin top-level-expr ...)

; module-level-expr is one of
;   general-top-level-expr
;   (provide provide-spec ...)
;   (begin module-level-expr ...)

; general-top-level-expr is one of
;   expr
;   (define-values (variable ...) expr)
;   (define-syntaxes (variable ...) expr)
;   (require require-spec ...)
;   (require-for-syntax require-spec ...)

; expr is one of
;   variable
;   (lambda formals expr ...1)
;   (case-lambda (formals expr ...1) ...)
;   (if expr expr)
;   (if expr expr expr)
;   (begin expr ...1)
;   (begin0 expr expr ...)
;   (let-values (((variable ...) expr) ...) expr ...1)
;   (letrec-values (((variable ...) expr) ...) expr ...1)
;   (set! variable expr)
;   (quote datum)
;   (quote-syntax datum)
;   (with-continuation-mark expr expr expr)
;   (#%app expr ...1)
;   (#%datum . datum)
;   (#%top . variable)

; reexpand : phase module-name expanded-module -> syntax-object
;  given an expanded module represented as the syntax-object stx,
;  return the "reexpansion" of the module, by renaming all identifiers to
;  include module-name, phase and previous name.
(define (reexpand p n stx)
  (reexpand-top-level-expr p 0 n stx))

; reexpand-top-level-expr : phase relative-phase name expanded-top-level-expr -> syntax-object
(define (reexpand-top-level-expr p rp n stx)
  (syntax-case* stx (module begin #%plain-module-begin) module-identifier=?
    [(module identifier name (#%plain-module-begin module-level-expr ...))
     #`(module identifier name
         (#%plain-module-begin
          #,@(reexpand-module-level-exprs p rp n #'(module-level-expr ...))))]
    [(begin top-level-expr ...)
     #`(begin #,@(reexpand-top-level-exprs p rp n #'(top-level-expr ...)))]
    [_
     (reexpand-general-top-level-expr p rp n stx)]))


(define (reexpand-top-level-exprs p rp n stx)
  (datum->syntax-object stx
                        (map (lambda (s) (reexpand-top-level-expr p rp n s))
                             (syntax->list stx))))


(define (reexpand-module-level-expr p rp n stx)
  (syntax-case stx (provide begin)
    [(provide provide-spec ...)
     #`(provide provide-spec ...)]
    [(begin module-level-expr ...)
     #`(begin #,(reexpand-module-level-exprs p rp n #'(module-level-expr ...)))]
    [_
     (reexpand-general-top-level-expr p rp n stx)]))


(define (reexpand-module-level-exprs p rp n stx)
  (datum->syntax-object stx
                        (map (lambda (s) (reexpand-module-level-expr p rp n s))
                             (syntax->list stx))))


(define (reexpand-general-top-level-expr p rp n stx)
  (define (g id-stx)   (generate-identifier-name p rp n id-stx))
  (define (gs ids-stx) (map g (syntax->list ids-stx)))
  (syntax-case stx (define-values define-syntaxes require require-for-syntax)    
    [(define-values (variable ...) expr)
     #`(define-values #,(gs #'(variable ...)) #,(reexpand-expr p rp n #'expr))]
    [(define-syntaxes (variable ...) expr)
     #`(define-syntaxes #,(gs #'(variable ...)) #,(reexpand-expr p (+ rp 1) n #'expr))]
    [(require require-spec ...)
     #`'(require require-spec ...)]
    [(require-for-syntax require-spec ...)
     #`'(require-for-syntax require-spec ...)]
    [_
     (reexpand-expr p rp n stx)]))

(define (reexpand-expr p rp n stx)
  (define (g id-stx)   (generate-identifier-name p rp n id-stx))
  (define (gs ids-stx) (map g (syntax->list ids-stx)))
  ;(for-each display (list "expr> " p rp n stx "\n"))
  (syntax-case* stx (lambda case-lambda if begin begin0 
                      let-values letrec-values set! quote quote-syntax 
                      with-continuation-mark #%app #%datum #%top)  (if (= rp 0)
                                                                       module-identifier=?
                                                                       module-transformer-identifier=?)
    [(lambda formals expr0 expr ...)
     #`(lambda #,(gs #'formals) #,@(reexpand-exprs p rp n #'(expr0 expr ...)))]
    [(case-lambda (formals expr0 expr ...) ...)
     #`(case-lambda #,@(reexpand-case-lambda-clauses p rp n #'((formals expr0 expr ...) ...)))]
    [(if expr0 expr1)
     #`(if #,(reexpand-expr p rp n #'expr0) #,(reexpand-expr p rp n #'expr1))]
    [(if expr0 expr1 expr2)
     #`(if #,(reexpand-expr p rp n #'expr0) #,(reexpand-expr p rp n #'expr1) #,(reexpand-expr p rp n #'expr2))]
    [(begin expr0 expr ...)
     #`(begin #,@(reexpand-exprs p rp n #'(expr0 expr ...)))]
    [(begin0 expr0 expr ...)
     #`(begin0 #,@(reexpand-exprs p rp n #'(expr0 expr ...)))]
    [(let-values (((variable ...) expr) ...) expr0 expr1 ...)
     #`(let-values #,(reexpand-let-values-clauses p rp n #'(((variable ...) expr) ...)) 
         #,@(reexpand-exprs p rp n #'(expr0 expr1 ...)))]
    [(letrec-values (((variable ...) expr) ...) expr0 expr1 ...)
     #`(letrec-values  #,(reexpand-letrec-values-clauses p rp n #'(((variable ...) expr) ...)) 
         #,@(reexpand-exprs p rp n #'(expr0 expr1 ...)))]
    [(set! variable expr)
     #`(set! #,(g #'variable) #,(reexpand-expr p rp n #'expr))]
    [(quote datum)
     #'(quote datum)]
    ; quote-syntax is the same as syntax without pattern variables
    [(quote-syntax datum)
     #`(quote-syntax #,(reexpand-quote-syntax p rp n #'datum))]
    [(with-continuation-mark expr0 expr1 expr2)
     #`(with-continuation-mark 
           #,(reexpand-expr p rp n #'expr0) #,(reexpand-expr p rp n #'expr1) #,(reexpand-expr p rp n #'expr))]
    [(#%app expr1 ...)
     #`(      #,@(reexpand-exprs p rp n #'(expr1 ...)))]
    #;[(#%app expr0 expr1 ...)
     (if *MZMODE*
         #`(#%app #,@(reexpand-exprs p rp n #'(expr0 expr1 ...)))
         #`(      #,@(reexpand-exprs p rp n #'(expr0 expr1 ...))))]
    [(#%datum . datum)
     (if *MZMODE* #'(#%datum . datum) #'datum)]
    [(#%top . variable)
     (if *MZMODE* #`(#%top . #,(g #'variable)) (g #'variable))]
    [variable
     (identifier? #'variable)
     (g #'variable)]
    [failed
     (begin 
       (display "Hprm!") (display #'failed) (newline)
       (error "huh?, got: " #'foo))]))

(define (reexpand-let-values-clauses p rp n clauses-stx)
  (define (g id-stx)   (generate-identifier-name p rp n id-stx))
  (define (gs ids-stx) (map g (syntax->list ids-stx)))
  (define (reexpand-clause clause)
    (syntax-case clause ()
      [((variable ...) expr)
       #`(#,(gs #'(variable ...)) #,(reexpand-expr p rp n #'expr))]))
  
  (datum->syntax-object clauses-stx (map reexpand-clause (syntax->list clauses-stx))))

(define reexpand-letrec-values-clauses reexpand-let-values-clauses)
 
;A case-lambda expression has the form: 
;(case-lambda  (formals expr ...1) ...)
;
;formals is one of 
;  variable
;  (variable ...)
;  (variable ... . variable)

;    [(case-lambda (formals expr0 expr ...) ...)
;     #`(case-lambda #,@(reexpand-case-lambda-clauses p rp n #'((formals expr0 expr ...) ...)))]

(define (reexpand-case-lambda-clauses p rp n clauses-stx)
  (define (g id-stx)   (generate-identifier-name p rp n id-stx))
  (define (gs ids-stx) (map g (syntax->list ids-stx)))
  (define (reexpand-clause clause)
    (syntax-case clause ()
      [((variable ...)       expr0 expr ...)   #`(#,(gs #'(variable ...)) #,@(reexpand-exprs p rp n #'(expr0 expr ...)))]
      [((variable ... . var) expr0 expr ...)   #`((#,@(gs #'(variable ...)) . #,(g #'var))
						  #,@(reexpand-exprs p rp n #'(expr0 expr ...)))]
      [(variable             expr0 expr ...)   #`(#,(g #'variable)        #,@(reexpand-exprs p rp n #'(expr0 expr ...)))]))

  (datum->syntax-object clauses-stx (map reexpand-clause (syntax->list clauses-stx))))
  
         
         
(define (reexpand-quote-syntax p rp n datum)
  (define (r d) (reexpand-quote-syntax p rp n d))
  (syntax-case datum ()
    [(a . d)                       #`(#,(r #'a) . #,(r #'d))]
    [id       (identifier? #'id)   (generate-identifier-name p rp n #'id)]
    [else                          datum]))
    
    
(define (reexpand-exprs p rp n stx)
  (datum->syntax-object stx
                        (map (lambda (s) (reexpand-expr p rp n s))
                             (syntax->list stx))))


;(define mod-A '
;  (module A mzscheme
;    (provide x inc-x! mx)
;    (define x 0)
;    (define-syntax mx (lambda (stx) #'x))
;    (define (inc-x!) (set! x (+ x 1)))))
;
;(define mod-B '
;  (module B mzscheme
;    (require A)
;    (inc-x!)
;    (display (mx)) (newline)))
;
;(define mod-C '
;  (module C mzscheme
;    (provide mx2)
;    (require A)
;    (require B)
;    (require-for-syntax A)
;    (define-syntax mx2 (lambda (stx) (datum->syntax-object stx (mx))))
;    (display "phase 1: ") (display (mx2)) (newline)
;    (display "phase 0: ") (display (mx))  (newline)))
;
;(eval mod-A)
;(eval mod-B)
;(eval mod-C)
;
;mod-A
;(reexpand 0 'a (expand mod-A))
;mod-B
;(reexpand 0 'b (expand mod-B))
;mod-C
;(reexpand 0 'c (expand mod-C))

;(trace parse-expanded
;       parse-top-level-expr
;       parse-module-level-expr
;       parse-general-top-level-expr
;       parse-expr)

;(define A-sexp
;  '(module A mzscheme
;     (provide x)
;     (define x 1)))
;
;

;(pretty-print
; (syntax-object->datum
;  (reexpand 0 'c (expand (read-syntax "c.scm" (open-input-file "c.scm"))))))


; CURRENT PROBLEM: the (require (lib "list.ss" mzlib")) in b.scm brings in spidey.ss
;                  which expands to something with source #%stxcase, and
;                  we end up with an error in generate-identifier.

; solution: introduced (syntax-case* .... (if (= rp 0) module-identifier=? module-transformer-identifier=?))
;           => now fixing left out reexpanders

; cause: something to do with #%stxcase being a kernel module?


; CURRENT MINOR PROBLEM:  Make sure that reexpand-case-lambda is correct.


;;; -------

(define (kernel-module-name? n)
  ; TODO: check #% names too
  (eq? n 'mzscheme))

(let ([reexpanded-modules '()])
  (process-dependencies 0 "c.scm" 
                        (lambda (p n stx)
                          (if (not (kernel-module-name? n))
                              (let ((reexpanded (reexpand p n stx)))
                                (set! reexpanded-modules 
                                      (cons ;(list p n reexpanded)
                                       (emit reexpanded)
                                       ;reexpanded
                                       reexpanded-modules))))))
  (for-each (lambda (o) 
              (pretty-print (syntax-object->datum o))
              (newline) (newline))
            (reverse reexpanded-modules)))
